from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.participant import ParticipantLayer
from models.models import PRIMARY_EXTERNAL_ORG, ParticipantUpsertInternal


class TestParticipant(DbIsolatedTest):
    """Test getting participants"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.ex01 = {PRIMARY_EXTERNAL_ORG: 'EX01', 'other': 'OTHER1'}
        self.ex02 = {PRIMARY_EXTERNAL_ORG: 'EX02'}

        pl = ParticipantLayer(self.connection)
        await pl.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids=self.ex01,
                    reported_sex=2,
                    karyotype='XX',
                    meta={'field': 1},
                ),
                ParticipantUpsertInternal(
                    external_ids=self.ex02,
                    reported_sex=1,
                    karyotype='XY',
                    meta={'field': 2},
                ),
            ]
        )

    @run_as_sync
    async def test_get_all_participants(self):
        """Test getting all participants"""
        pl = ParticipantLayer(self.connection)
        ps = await pl.get_participants(project=1)

        self.assertEqual(2, len(ps))

        self.assertEqual(self.ex01, ps[0].external_ids)
        self.assertEqual(1, ps[0].meta['field'])
        self.assertEqual('XX', ps[0].karyotype)

        self.assertEqual(self.ex02, ps[1].external_ids)

    @run_as_sync
    async def test_get_participant_by_eid(self):
        """Test to see what's in the database"""
        pl = ParticipantLayer(self.connection)
        ps = await pl.get_participants(project=1, external_participant_ids=['EX02'])

        self.assertEqual(1, len(ps))

        self.assertEqual(self.ex02, ps[0].external_ids)
        self.assertEqual(2, ps[0].meta['field'])
        self.assertEqual('XY', ps[0].karyotype)



from test.testbase import DbIsolatedTest, run_as_sync

from graphql.error import GraphQLError, GraphQLSyntaxError

import api.graphql.schema
from db.python.layers import AnalysisLayer, ParticipantLayer
from db.python.layers.family import FamilyLayer
from metamist.graphql import configure_sync_client, gql, validate
from models.enums import AnalysisStatus
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AnalysisInternal,
    AssayUpsertInternal,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.utils.sequencing_group_id_format import sequencing_group_id_format

default_assay_meta = {
    'sequencing_type': 'genome',
    'sequencing_technology': 'short-read',
    'sequencing_platform': 'illumina',
}


def _get_single_participant_upsert():
    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                meta={},
                type='blood',
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                    **default_assay_meta,
                                },
                            ),
                        ],
                    )
                ],
            )
        ],
    )


TEST_QUERY = gql(
    """
query MyQuery($project: String!) {
  project(name: $project) {
    participants {
      id
      samples {
        id
        sequencingGroups {
          id
          assays {
            id
          }
        }
      }
    }
  }
}"""
)


class TestGraphQL(DbIsolatedTest):
    """Test graphql functionality"""

    @run_as_sync
    async def setUp(self) -> None:
        """Setup the tests"""
        super().setUp()
        self.player = ParticipantLayer(self.connection)

    def test_validate_local_schema(self):
        """
        test using the bundled schema file (from regenerateapi.py)
         to make sure people can validate with authentication
        """
        validate(TEST_QUERY, use_local_schema=True)

    def test_validate_provided_schema(self):
        """
        Validate using schema directly from api.graphql.schema
        (strawberry has an as_str() method)
        """
        client = configure_sync_client(
            schema=api.graphql.schema.schema.as_str(),
            auth_token='FAKE',  # type: ignore
        )
        validate(TEST_QUERY, client=client)

    def test_bad_syntax_query(self):
        """Fail on bad syntax"""
        with self.assertRaises(GraphQLSyntaxError):
            gql(
                """
            query MyQuery(badtoken $project: String!) {
                project(name: $project) {
                    name
                }
            }"""
            )

    def test_bad_field_query(self):
        """Fail because the field doesn't exist"""
        # query syntactically validates
        query = gql(
            """
            query MyQuery($project: String!) {
                project(name: $project) {
                    thisFieldDoesntExist
                }
        }"""
        )
        with self.assertRaises(GraphQLError):
            validate(query, use_local_schema=True)

    @run_as_sync
    async def test_basic_graphql_query(self):
        """Test getting the summary for a project"""
        p = (await self.player.upsert_participants([_get_single_participant_upsert()]))[
            0
        ]

        query = """
query MyQuery($project: String!) {
  project(name: $project) {
    participants {
      id
      samples {
        id
        sequencingGroups {
          id
          assays {
            id
          }
        }
      }
    }
  }
}"""
        data = await self.run_graphql_query_async(
            query, variables={'project': self.project_name}
        )
        participants = data['project']['participants']
        self.assertEqual(1, len(participants))
        self.assertEqual(p.id, participants[0]['id'])
        samples = participants[0]['samples']
        self.assertEqual(1, len(samples))

        self.assertEqual(p.samples[0].to_external().id, samples[0]['id'])
        sequencing_groups = samples[0]['sequencingGroups']
        self.assertEqual(1, len(sequencing_groups))
        self.assertEqual(
            p.samples[0].sequencing_groups[0].to_external().id,
            sequencing_groups[0]['id'],
        )
        assays = sequencing_groups[0]['assays']
        self.assertEqual(
            1, len(participants[0]['samples'][0]['sequencingGroups'][0]['assays'])
        )
        self.assertEqual(
            p.samples[0].sequencing_groups[0].assays[0].id, assays[0]['id']
        )

    @run_as_sync
    async def test_query_sample_by_meta(self):
        """Test querying a participant"""
        await self.player.upsert_participant(
            ParticipantUpsertInternal(
                meta={},
                external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
                samples=[
                    SampleUpsertInternal(
                        external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                        meta={'thisKey': 'value'},
                    )
                ],
            )
        )
        q = """
    query MyQuery($project: String!, $meta: JSON!) {
        project(name: $project) {
            participants {
                samples(meta: $meta) {
                    id
                }
            }
        }
    }"""
        values = await self.run_graphql_query_async(
            q, {'project': self.project_name, 'meta': {'thisKey': 'value'}}
        )
        assert values

        self.assertEqual(1, len(values['project']['participants'][0]['samples']))

        values2 = await self.run_graphql_query_async(
            q, {'project': self.project_name, 'meta': {'thisKeyDoesNotExistEver': '-1'}}
        )
        assert values2

        self.assertEqual(0, len(values2['project']['participants'][0]['samples']))

    @run_as_sync
    async def test_sg_analyses_query(self):
        """Example graphql query of analyses from sequencing-group"""
        p = await self.player.upsert_participant(_get_single_participant_upsert())
        sg_id = p.samples[0].sequencing_groups[0].id

        alayer = AnalysisLayer(self.connection)
        await alayer.create_analysis(
            AnalysisInternal(
                sequencing_group_ids=[sg_id],
                type='cram',
                status=AnalysisStatus.COMPLETED,
                meta={},
                output='some-output',
            )
        )

        q = """
query MyQuery($sg_id: String!, $project: String!) {
  sequencingGroups(id: {in_: [$sg_id]}, project: {eq: $project}) {
    analyses(project: {eq: $project}) {
      id
      meta
      output
    }
  }
}"""

        resp = await self.run_graphql_query_async(
            q,
            {'sg_id': sequencing_group_id_format(sg_id), 'project': self.project_name},
        )
        self.assertIn('sequencingGroups', resp)
        self.assertEqual(1, len(resp['sequencingGroups']))
        self.assertIn('analyses', resp['sequencingGroups'][0])
        self.assertEqual(1, len(resp['sequencingGroups'][0]['analyses']))
        analyses = resp['sequencingGroups'][0]['analyses']
        self.assertIn('id', analyses[0])
        self.assertIn('meta', analyses[0])
        self.assertIn('output', analyses[0])

    @run_as_sync
    async def test_participant_phenotypes(self):
        """
        Test getting participant phentypes in graphql
        """
        # insert participant
        p = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'}, meta={}, samples=[]
            )
        )

        phenotypes = {'phenotype1': 'value1', 'phenotype2': {'number': 123}}
        # insert participant_phenotypes
        await self.player.insert_participant_phenotypes({p.id: phenotypes})

        q = """
query MyQuery($pid: Int!) {
  participant(id: $pid) {
    phenotypes
  }
}"""

        resp = await self.run_graphql_query_async(q, {'pid': p.id})

        self.assertIn('participant', resp)
        self.assertIn('phenotypes', resp['participant'])
        self.assertDictEqual(phenotypes, resp['participant']['phenotypes'])

    @run_as_sync
    async def test_family_participants(self):
        """Test inserting + querying family participants from different directions"""
        family_layer = FamilyLayer(self.connection)

        family_eid = 'family1'

        rows = [
            [family_eid, 'individual1', 'paternal1', 'maternal1', 'm', '1', 'note1'],
            [family_eid, 'paternal1', None, None, 'm', '0', 'note2'],
            [family_eid, 'maternal1', None, None, 'f', '1', 'note3'],
        ]

        await family_layer.import_pedigree(None, rows, create_missing_participants=True)

        q = """
query MyQuery($project: String!) {
    project(name: $project) {
        participants {
            externalId
            familyParticipants {
                affected
                notes
                family {
                    externalId
                }
            }
            families {
                externalId
            }
        }
        families {
            externalId
            familyParticipants {
                affected
                notes
                participant {
                    externalId
                }
            }
        }
    }
}
"""

        resp = await self.run_graphql_query_async(q, {'project': self.project_name})
        assert resp is not None

        family_simple_obj = {'family': {'externalId': family_eid}}

        participants = resp['project']['participants']
        families = resp['project']['families']

        participants_by_eid = {p['externalId']: p for p in participants}
        self.assertEqual(3, len(participants))

        self.assertDictEqual(
            {
                'externalId': 'individual1',
                'families': [{'externalId': family_eid}],
                'familyParticipants': [
                    {'affected': 1, 'notes': 'note1', **family_simple_obj}
                ],
            },
            participants_by_eid['individual1'],
        )
        self.assertEqual(1, len(participants_by_eid['individual1']['families']))

        self.assertEqual(1, len(families))
        self.assertEqual(family_eid, families[0]['externalId'])

        sorted_fps = sorted(
            families[0]['familyParticipants'],
            key=lambda x: x['participant']['externalId'],
        )
        self.assertListEqual(
            sorted_fps,
            [
                {
                    'affected': 1,
                    'notes': 'note1',
                    'participant': {'externalId': 'individual1'},
                },
                {
                    'affected': 1,
                    'notes': 'note3',
                    'participant': {'externalId': 'maternal1'},
                },
                {
                    'affected': 0,
                    'notes': 'note2',
                    'participant': {'externalId': 'paternal1'},
                },
            ],
        )

    @run_as_sync
    async def test_get_project_name_from_analysis(self):
        """Test getting project name from analysis"""
        p = await self.player.upsert_participant(_get_single_participant_upsert())
        sg_id = p.samples[0].sequencing_groups[0].id

        alayer = AnalysisLayer(self.connection)
        await alayer.create_analysis(
            AnalysisInternal(
                sequencing_group_ids=[sg_id],
                type='cram',
                status=AnalysisStatus.COMPLETED,
                meta={},
                output='some-output',
            )
        )

        q = """
query MyQuery($sg_id: String!) {
  sequencingGroups(id: {eq: $sg_id}) {
    analyses {
      id
      project {
        name
      }
    }
  }
}"""

        resp = await self.run_graphql_query_async(
            q, {'sg_id': sequencing_group_id_format(sg_id)}
        )
        self.assertIn('sequencingGroups', resp)
        project_name = resp['sequencingGroups'][0]['analyses'][0]['project']['name']
        self.assertEqual(self.project_name, project_name)



from collections import defaultdict
from datetime import datetime
from math import ceil
from test.testbase import DbIsolatedTest, run_as_sync

from db.python.enum_tables.sample_type import SampleTypeTable
from db.python.layers.ourdna.dashboard import OurDnaDashboardLayer
from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    OurDNADashboard,
    ParticipantUpsertInternal,
    SampleUpsert,
    SampleUpsertInternal,
)


def str_to_datetime(timestamp_str):
    """Convert string timestamp to datetime"""
    return datetime.strptime(timestamp_str, '%Y-%m-%d %H:%M:%S')


class OurDNADashboardTest(DbIsolatedTest):
    """Test sample class"""

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()
        self.odd = OurDnaDashboardLayer(self.connection)
        self.sl = SampleLayer(self.connection)
        self.pl = ParticipantLayer(self.connection)

        self.stt = SampleTypeTable(self.connection)
        await self.stt.insert('ebld')

        participants = await self.pl.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'},
                    reported_sex=2,
                    karyotype='XX',
                    meta={'consent': True, 'field': 1},
                    samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                            meta={
                                'collection-time': '2022-07-03 13:28:00',
                                'processing-site': 'Garvan',
                                'process-start-time': '2022-07-06 16:28:00',
                                'process-end-time': '2022-07-06 19:28:00',
                                'received-time': '2022-07-03 14:28:00',
                                'received-by': 'YP',
                                'collection-lab': 'XYZ LAB',
                                'collection-event-type': 'walk-in',
                                'courier': 'ABC COURIERS',
                                'courier-tracking-number': 'ABCDEF12562',
                                'courier-scheduled-pickup-time': '2022-07-03 13:28:00',
                                'courier-actual-pickup-time': '2022-07-03 13:28:00',
                                'courier-scheduled-dropoff-time': '2022-07-03 13:28:00',
                                'courier-actual-dropoff-time': '2022-07-03 13:28:00',
                                'concentration': 1.45,
                            },
                            type='ebld',
                            active=True,
                            nested_samples=[
                                # add a random sample here to test it's not collected
                                SampleUpsertInternal(
                                    external_ids={PRIMARY_EXTERNAL_ORG: 'Test01-01'},
                                    type='blood',
                                    meta={
                                        # something wild
                                        'collection-time': '1999-01-01 12:34:56',
                                    },
                                )
                            ],
                        )
                    ],
                ),
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX02'},
                    reported_sex=1,
                    karyotype='XY',
                    meta={'field': 2},
                    samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Test02'},
                            meta={
                                'collection-time': '2022-07-03 13:28:00',
                                'processing-site': 'BBV',
                                # 'process-start-time': '2022-07-06 16:28:00',
                                # 'process-end-time': '2022-07-06 19:28:00',
                                'received-time': '2022-07-03 14:28:00',
                                'received-by': 'YP',
                                'collection-lab': 'XYZ LAB',
                                'collection-event-type': 'OSS',
                                'courier': 'ABC COURIERS',
                                'courier-tracking-number': 'ABCDEF12562',
                                'courier-scheduled-pickup-time': '2022-07-03 13:28:00',
                                'courier-actual-pickup-time': '2022-07-03 13:28:00',
                                'courier-scheduled-dropoff-time': '2022-07-03 13:28:00',
                                'courier-actual-dropoff-time': '2022-07-03 13:28:00',
                                'concentration': 0.98,
                            },
                            type='ebld',
                            active=True,
                        )
                    ],
                ),
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX03'},
                    reported_sex=2,
                    karyotype='XX',
                    meta={'consent': True, 'field': 3},
                    samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'Test03'},
                            meta={
                                # 'collection-time': '2022-07-03 13:28:00',
                                'processing-site': 'Garvan',
                                # 'process-start-time': '2022-07-03 16:28:00',
                                # 'process-end-time': '2022-07-03 19:28:00',
                                'received-time': '2022-07-03 14:28:00',
                                'received-by': 'YP',
                                'collection-lab': 'XYZ LAB',
                                'courier': 'ABC COURIERS',
                                'courier-tracking-number': 'ABCDEF12562',
                                'courier-scheduled-pickup-time': '2022-07-03 13:28:00',
                                'courier-actual-pickup-time': '2022-07-03 13:28:00',
                                'courier-scheduled-dropoff-time': '2022-07-03 13:28:00',
                                'courier-actual-dropoff-time': '2022-07-03 13:28:00',
                                'concentration': 1.66,
                            },
                            type='ebld',
                            active=True,
                        )
                    ],
                ),
            ],
        )

        self.participants_external_objects = [
            participant.to_external() for participant in participants
        ]

        self.sample_external_objects: list[SampleUpsert] = []
        self.sample_internal_objects: list[SampleUpsertInternal] = []

        self.sample_internal_objects.extend(
            sample for participant in participants for sample in participant.samples
        )
        self.sample_external_objects.extend(
            sample.to_external() for sample in self.sample_internal_objects
        )

    @run_as_sync
    async def test_get_dashboard(self):
        """Test get_dashboard"""
        dashboard = await self.odd.query(project_id=self.project_id)

        # Check that the dashboard is not empty and is a dict
        self.assertTrue(dashboard)
        self.assertIsInstance(dashboard, OurDNADashboard)

    @run_as_sync
    async def test_collection_to_process_end_time(self):
        """I want to know how long it took between blood collection and sample processing"""
        dashboard = await self.odd.query(project_id=self.project_id)
        collection_to_process_end_time = dashboard.collection_to_process_end_time

        # Check that collection_to_process_end_time is not empty and is a dict
        self.assertTrue(collection_to_process_end_time)
        self.assertIsInstance(collection_to_process_end_time, dict)

        samples_filtered: list[str] = []
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            collection_time = sample.meta.get('collection-time')
            process_end_time = sample.meta.get('process-end-time')
            # Skip samples that don't have collection_time or process_end_time
            if not collection_time or not process_end_time:
                continue
            time_difference = str_to_datetime(process_end_time) - str_to_datetime(
                collection_time
            )
            if time_difference.total_seconds():
                # Check that the time difference matches
                assert isinstance(sample.id, str)

                samples_filtered.append(sample.id)

                self.assertEqual(
                    time_difference.total_seconds(),
                    collection_to_process_end_time[sample.id],
                )

        # Check the number of samples in the cohort
        self.assertCountEqual(collection_to_process_end_time.keys(), samples_filtered)

        for _sample_id, time_diff in collection_to_process_end_time.items():
            self.assertGreater(time_diff, 0)
            self.assertIsInstance(time_diff, int)

    @run_as_sync
    async def test_collection_to_process_end_time_24h(self):
        """I want to know which samples took more than 24 hours between blood collection and sample processing completion"""
        dashboard = await self.odd.query(project_id=self.project_id)
        collection_to_process_end_time_24h = (
            dashboard.collection_to_process_end_time_24h
        )

        # Check that collection_to_process_end_time is not empty and is a dict
        self.assertTrue(collection_to_process_end_time_24h)
        self.assertIsInstance(collection_to_process_end_time_24h, dict)

        samples_filtered: list[str] = []
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            collection_time = sample.meta.get('collection-time')
            process_end_time = sample.meta.get('process-end-time')
            # Skip samples that don't have collection_time or process_end_time
            if not collection_time or not process_end_time:
                continue
            time_difference = str_to_datetime(process_end_time) - str_to_datetime(
                collection_time
            )
            if time_difference.total_seconds() > 24 * 3600:
                # Check that the time difference matches
                assert isinstance(sample.id, str)

                samples_filtered.append(sample.id)

                self.assertEqual(
                    time_difference.total_seconds(),
                    collection_to_process_end_time_24h[sample.id],
                )

        # check that there are a correct number of matching results
        self.assertCountEqual(
            collection_to_process_end_time_24h.keys(), samples_filtered
        )

    @run_as_sync
    async def test_processing_times_per_site(self):
        """I want to know what the sample processing times were for samples at each designated site"""
        dashboard = await self.odd.query(project_id=self.project_id)
        processing_times_by_site = dashboard.processing_times_by_site

        # Check that processing_times_per_site is not empty and is a dict
        self.assertTrue(processing_times_by_site)
        self.assertIsInstance(processing_times_by_site, dict)

        sample_tally: dict[str, dict[int, int]] = defaultdict()
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            processing_site = sample.meta.get('processing-site', 'Unknown')
            process_start_time = sample.meta.get('process-start-time')
            process_end_time = sample.meta.get('process-end-time')
            # Skip samples that don't have process_start_time or process_end_time
            if not process_start_time or not process_end_time:
                continue
            time_difference = str_to_datetime(process_end_time) - str_to_datetime(
                process_start_time
            )
            current_bucket = ceil(time_difference.total_seconds() / 3600)
            if processing_site in sample_tally:
                sample_tally[processing_site][current_bucket] += 1
            else:
                sample_tally[processing_site] = {}
                sample_tally[processing_site][current_bucket] = 1

        # Checks that we have identical dicts (by extension, keys and their values)
        self.assertDictEqual(processing_times_by_site, sample_tally)

    @run_as_sync
    async def test_total_samples_by_collection_event_name(self):
        """I want to know how many samples were collected from walk-ins vs during events or scheduled activities"""
        dashboard = await self.odd.query(project_id=self.project_id)
        total_samples_by_collection_event_name = (
            dashboard.total_samples_by_collection_event_name
        )

        # Check that total_samples_by_collection_event_name is not empty and is a dict
        self.assertTrue(total_samples_by_collection_event_name)
        self.assertIsInstance(total_samples_by_collection_event_name, dict)

        sample_tally: dict[str, int] = defaultdict()
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            event_name = sample.meta.get('collection-event-type', 'Unknown')
            if event_name in sample_tally:
                sample_tally[event_name] += 1
            else:
                sample_tally[event_name] = 1

        # Check that the tally and the total_samples_by_collection_event_name are the same, by extension, keys and their values
        self.assertDictEqual(total_samples_by_collection_event_name, sample_tally)

    @run_as_sync
    async def test_samples_lost_after_collection(self):
        """I need to know how many samples have been lost, EG: participants have been consented, blood collected, not processed"""
        dashboard = await self.odd.query(project_id=self.project_id)
        samples_lost_after_collection = dashboard.samples_lost_after_collection

        # Check that samples_lost_after_collection is not empty and is a dict
        self.assertTrue(samples_lost_after_collection)
        self.assertIsInstance(samples_lost_after_collection, list)

        # Check that the number of samples in the list is correct
        samples_filtered: list[str] = []
        sample_ids_lost_after_collection = [
            sample.sample_id for sample in samples_lost_after_collection
        ]
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            collection_time = sample.meta.get('collection-time')
            process_start_time = sample.meta.get('process-start-time')
            # Skip samples that don't have collection_time or process_end_time
            if not collection_time:
                continue
            time_difference = datetime.now() - str_to_datetime(collection_time)
            if time_difference.total_seconds() > 72 * 3600 and not process_start_time:
                # Check that the time difference matches
                assert isinstance(sample.id, str)

                samples_filtered.append(sample.id)

                for sample_data in samples_lost_after_collection:
                    if sample_data.sample_id == sample.id:
                        self.assertEqual(
                            int(time_difference.total_seconds()),
                            sample_data.time_since_collection,
                        )

        # check that there are a correct number of matching results
        self.assertCountEqual(sample_ids_lost_after_collection, samples_filtered)

        # TODO: Add assertions VB

    @run_as_sync
    async def test_samples_more_than_1ug_dna(self):
        """I want to generate a list of samples containing more than 1 ug of DNA to prioritise them for long-read sequencing applications"""
        dashboard = await self.odd.query(project_id=self.project_id)
        samples_more_than_1ug_dna = dashboard.samples_concentration_gt_1ug

        # Check that samples_concentratiom_gt_1ug is not empty and is a dict
        self.assertTrue(samples_more_than_1ug_dna)
        self.assertIsInstance(samples_more_than_1ug_dna, dict)

        # Check that the number of samples in the list is correct
        samples_filtered: list[str] = []
        for sample in self.sample_external_objects:
            assert isinstance(sample.meta, dict)
            if sample.meta['concentration'] and sample.meta['concentration'] > 1:
                assert isinstance(sample.id, str)

                samples_filtered.append(sample.id)

        self.assertCountEqual(samples_more_than_1ug_dna, samples_filtered)

    @run_as_sync
    async def test_participants_consented_not_collected(self):
        """I want to know how many people who have consented and NOT given blood"""
        dashboard = await self.odd.query(project_id=self.project_id)
        # print(dashboard)
        participants_consented_not_collected = (
            dashboard.participants_consented_not_collected
        )

        # Check that participants_consented_not_collected is not empty and is a dict
        self.assertTrue(participants_consented_not_collected)
        self.assertIsInstance(participants_consented_not_collected, list)

        # Check that the number of participants in the list is correct
        participants_filtered: list[int] = []
        for participant in self.participants_external_objects:
            assert isinstance(participant.meta, dict)
            samples_for_participant = [
                sample
                for sample in self.sample_external_objects
                if sample.participant_id == participant.id
                and isinstance(sample.meta, dict)
            ]
            if participant.meta.get('consent') and any(
                isinstance(sample.meta, dict)
                and sample.meta.get('collection-time') is None
                for sample in samples_for_participant
            ):
                assert isinstance(participant.id, int)
                participants_filtered.append(participant.id)

        self.assertCountEqual(
            participants_consented_not_collected, participants_filtered
        )

    @run_as_sync
    async def test_participants_signed_not_consented(self):
        """I want to know how many people have signed up but not consented"""
        dashboard = await self.odd.query(project_id=self.project_id)
        # print(dashboard)
        participants_signed_not_consented = dashboard.participants_signed_not_consented

        # Check that participants_signed_not_consented is not empty and is a dict
        self.assertTrue(participants_signed_not_consented)
        self.assertIsInstance(participants_signed_not_consented, list)

        # Check that the number of participants in the list is correct
        participants_filtered: list[int] = []
        for participant in self.participants_external_objects:
            assert isinstance(participant.meta, dict)
            if not participant.meta.get('consent'):
                assert isinstance(participant.id, int)
                participants_filtered.append(participant.id)

        self.assertCountEqual(participants_signed_not_consented, participants_filtered)



# pylint: disable=invalid-overridden-method
import datetime
from test.testbase import DbIsolatedTest, run_as_sync

from db.python.filters import GenericFilter
from db.python.layers.analysis_runner import AnalysisRunnerLayer
from db.python.tables.analysis_runner import AnalysisRunnerFilter
from models.models.analysis_runner import AnalysisRunnerInternal


class TestAnalysisRunner(DbIsolatedTest):
    """Test sample class"""

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()

        self.al = AnalysisRunnerLayer(self.connection)

    @run_as_sync
    async def test_insert(self):
        """Test insert"""
        analysis = AnalysisRunnerInternal(
            ar_guid='<ar-guid>',
            project=self.project_id,
            output_path='output_path',
            timestamp=datetime.datetime(2024, 1, 1),
            access_level='test',
            repository='repository',
            config_path='config_path',
            environment='gcp',
            submitting_user='submitting_user',
            commit='commit',
            script='script',
            description='description',
            hail_version='1.0',
            cwd='cwd',
            driver_image='driver_image',
            batch_url='batch_url',
            meta={'meta': 'meta'},
        )
        await self.al.insert_analysis_runner_entry(analysis)

        db_ars = await self.al.query(
            AnalysisRunnerFilter(ar_guid=GenericFilter(eq=analysis.ar_guid))
        )
        self.assertEqual(len(db_ars), 1)
        field_to_compare = [
            'project',
            'output_path',
            'access_level',
            'repository',
            'config_path',
            'environment',
            'submitting_user',
            'commit',
            'script',
            'description',
            'hail_version',
            'cwd',
            'driver_image',
            'batch_url',
            'meta',
        ]
        for field in field_to_compare:
            # check each field is the same
            self.assertEqual(
                getattr(db_ars[0], field),
                getattr(analysis, field),
                msg='Field: ' + field,
            )

    @run_as_sync
    async def test_query(self):
        """
        Query all the Filter fields to check they work correctly
        """
        analyses = [
            AnalysisRunnerInternal(
                ar_guid=f'<ar-guid-{i+1}>',
                project=self.project_id,
                output_path='output_path',
                timestamp=datetime.datetime(2024, 1, 1),
                access_level='test',
                repository='repository',
                config_path='config_path',
                environment='gcp',
                submitting_user='submitting_user',
                commit='commit',
                script='script',
                description='description',
                hail_version='1.0',
                cwd='cwd',
                driver_image='driver_image',
                batch_url='batch_url',
                meta={'meta': 'meta'},
            )
            for i in range(3)
        ]
        for analysis in analyses:
            await self.al.insert_analysis_runner_entry(analysis)

        db_ars = await self.al.query(
            AnalysisRunnerFilter(
                project=GenericFilter(eq=self.project_id),
                submitting_user=GenericFilter(eq='submitting_user'),
                repository=GenericFilter(eq='repository'),
                access_level=GenericFilter(eq='test'),
                environment=GenericFilter(eq='gcp'),
            )
        )

        # return all 3
        self.assertEqual(len(db_ars), 3)

        # get one for 2 of the ar-guids
        guids_to_query = {a.ar_guid for a in analyses[:2]}
        db_ars = await self.al.query(
            AnalysisRunnerFilter(
                ar_guid=GenericFilter(in_=list(guids_to_query)),
            )
        )
        self.assertEqual(len(db_ars), 2)
        self.assertSetEqual({a.ar_guid for a in db_ars}, guids_to_query)



"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 7.4.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from metamist.api_client import ApiClient, Endpoint as _Endpoint
from metamist.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types,
    async_wrap
)
from metamist.model.body_export_project_participants import BodyExportProjectParticipants
from metamist.model.export_type import ExportType
from metamist.model.http_validation_error import HTTPValidationError
from metamist.model.project_participant_grid_filter import ProjectParticipantGridFilter
from metamist.model.project_participant_grid_response import ProjectParticipantGridResponse
from metamist.model.project_summary import ProjectSummary
from metamist.model.search_response_model import SearchResponseModel
from metamist.model.seqr_dataset_type import SeqrDatasetType


class WebApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        self.export_project_participants_async = async_wrap(self.export_project_participants)
        self.export_project_participants_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/{project}/participants/export/{export_type}',
                'operation_id': 'export_project_participants',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'export_type',
                    'project',
                    'body_export_project_participants',
                ],
                'required': [
                    'export_type',
                    'project',
                    'body_export_project_participants',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'export_type':
                        (ExportType,),
                    'project':
                        (str,),
                    'body_export_project_participants':
                        (BodyExportProjectParticipants,),
                },
                'attribute_map': {
                    'export_type': 'export_type',
                    'project': 'project',
                },
                'location_map': {
                    'export_type': 'path',
                    'project': 'path',
                    'body_export_project_participants': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.get_project_participants_filter_schema_async = async_wrap(self.get_project_participants_filter_schema)
        self.get_project_participants_filter_schema_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/{project}/participants/schema',
                'operation_id': 'get_project_participants_filter_schema',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_project_participants_grid_with_limit_async = async_wrap(self.get_project_participants_grid_with_limit)
        self.get_project_participants_grid_with_limit_endpoint = _Endpoint(
            settings={
                'response_type': (ProjectParticipantGridResponse,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/{project}/participants',
                'operation_id': 'get_project_participants_grid_with_limit',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'limit',
                    'project_participant_grid_filter',
                    'skip',
                ],
                'required': [
                    'project',
                    'limit',
                    'project_participant_grid_filter',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'limit':
                        (int,),
                    'project_participant_grid_filter':
                        (ProjectParticipantGridFilter,),
                    'skip':
                        (int,),
                },
                'attribute_map': {
                    'project': 'project',
                    'limit': 'limit',
                    'skip': 'skip',
                },
                'location_map': {
                    'project': 'path',
                    'limit': 'query',
                    'project_participant_grid_filter': 'body',
                    'skip': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.get_project_summary_async = async_wrap(self.get_project_summary)
        self.get_project_summary_endpoint = _Endpoint(
            settings={
                'response_type': (ProjectSummary,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/{project}/summary',
                'operation_id': 'get_project_summary',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.search_by_keyword_async = async_wrap(self.search_by_keyword)
        self.search_by_keyword_endpoint = _Endpoint(
            settings={
                'response_type': (SearchResponseModel,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/search',
                'operation_id': 'search_by_keyword',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'keyword',
                ],
                'required': [
                    'keyword',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'keyword':
                        (str,),
                },
                'attribute_map': {
                    'keyword': 'keyword',
                },
                'location_map': {
                    'keyword': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.sync_seqr_project_async = async_wrap(self.sync_seqr_project)
        self.sync_seqr_project_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/web/{project}/{sequencing_type}/sync-dataset',
                'operation_id': 'sync_seqr_project',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'sequencing_type',
                    'project',
                    'seqr_dataset_type',
                    'sync_families',
                    'sync_individual_metadata',
                    'sync_individuals',
                    'sync_es_index',
                    'sync_saved_variants',
                    'sync_cram_map',
                    'post_slack_notification',
                ],
                'required': [
                    'sequencing_type',
                    'project',
                    'seqr_dataset_type',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'sequencing_type':
                        (str,),
                    'project':
                        (str,),
                    'seqr_dataset_type':
                        ([SeqrDatasetType],),
                    'sync_families':
                        (bool,),
                    'sync_individual_metadata':
                        (bool,),
                    'sync_individuals':
                        (bool,),
                    'sync_es_index':
                        (bool,),
                    'sync_saved_variants':
                        (bool,),
                    'sync_cram_map':
                        (bool,),
                    'post_slack_notification':
                        (bool,),
                },
                'attribute_map': {
                    'sequencing_type': 'sequencing_type',
                    'project': 'project',
                    'sync_families': 'sync_families',
                    'sync_individual_metadata': 'sync_individual_metadata',
                    'sync_individuals': 'sync_individuals',
                    'sync_es_index': 'sync_es_index',
                    'sync_saved_variants': 'sync_saved_variants',
                    'sync_cram_map': 'sync_cram_map',
                    'post_slack_notification': 'post_slack_notification',
                },
                'location_map': {
                    'sequencing_type': 'path',
                    'project': 'path',
                    'seqr_dataset_type': 'body',
                    'sync_families': 'query',
                    'sync_individual_metadata': 'query',
                    'sync_individuals': 'query',
                    'sync_es_index': 'query',
                    'sync_saved_variants': 'query',
                    'sync_cram_map': 'query',
                    'post_slack_notification': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

    def export_project_participants(
        self,
        export_type,
        project,
        body_export_project_participants,
        **kwargs
    ):
        """Export Project Participants  # noqa: E501

        Get project summary (from query) with some limit  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.export_project_participants(export_type, project, body_export_project_participants, async_req=True)
        >>> result = thread.get()

        Args:
            export_type (ExportType):
            project (str):
            body_export_project_participants (BodyExportProjectParticipants):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['export_type'] = \
            export_type
        kwargs['project'] = \
            project
        kwargs['body_export_project_participants'] = \
            body_export_project_participants
        return self.export_project_participants_endpoint.call_with_http_info(**kwargs)

    def get_project_participants_filter_schema(
        self,
        project,
        **kwargs
    ):
        """Get Project Project Participants Filter Schema  # noqa: E501

        Get project summary (from query) with some limit  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_project_participants_filter_schema(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.get_project_participants_filter_schema_endpoint.call_with_http_info(**kwargs)

    def get_project_participants_grid_with_limit(
        self,
        project,
        limit,
        project_participant_grid_filter,
        **kwargs
    ):
        """Get Project Participants Grid With Limit  # noqa: E501

        Get project summary (from query) with some limit  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_project_participants_grid_with_limit(project, limit, project_participant_grid_filter, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            limit (int):
            project_participant_grid_filter (ProjectParticipantGridFilter):

        Keyword Args:
            skip (int): [optional] if omitted the server will use the default value of 0
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            ProjectParticipantGridResponse
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['limit'] = \
            limit
        kwargs['project_participant_grid_filter'] = \
            project_participant_grid_filter
        return self.get_project_participants_grid_with_limit_endpoint.call_with_http_info(**kwargs)

    def get_project_summary(
        self,
        project,
        **kwargs
    ):
        """Get Project Summary  # noqa: E501

        Creates a new sample, and returns the internal sample ID  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_project_summary(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            ProjectSummary
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.get_project_summary_endpoint.call_with_http_info(**kwargs)

    def search_by_keyword(
        self,
        keyword,
        **kwargs
    ):
        """Search By Keyword  # noqa: E501

        This searches the keyword, in families, participants + samples in the projects that you are a part of (automatically).  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.search_by_keyword(keyword, async_req=True)
        >>> result = thread.get()

        Args:
            keyword (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            SearchResponseModel
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['keyword'] = \
            keyword
        return self.search_by_keyword_endpoint.call_with_http_info(**kwargs)

    def sync_seqr_project(
        self,
        sequencing_type,
        project,
        seqr_dataset_type,
        **kwargs
    ):
        """Sync Seqr Project  # noqa: E501

        Sync a metamist project with its seqr project (for a specific sequence type) es_index_types: list of any of 'Haplotypecaller', 'SV_Caller', 'Mitochondria_Caller'  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.sync_seqr_project(sequencing_type, project, seqr_dataset_type, async_req=True)
        >>> result = thread.get()

        Args:
            sequencing_type (str):
            project (str):
            seqr_dataset_type ([SeqrDatasetType]):

        Keyword Args:
            sync_families (bool): [optional] if omitted the server will use the default value of True
            sync_individual_metadata (bool): [optional] if omitted the server will use the default value of True
            sync_individuals (bool): [optional] if omitted the server will use the default value of True
            sync_es_index (bool): [optional] if omitted the server will use the default value of True
            sync_saved_variants (bool): [optional] if omitted the server will use the default value of True
            sync_cram_map (bool): [optional] if omitted the server will use the default value of True
            post_slack_notification (bool): [optional] if omitted the server will use the default value of True
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['sequencing_type'] = \
            sequencing_type
        kwargs['project'] = \
            project
        kwargs['seqr_dataset_type'] = \
            seqr_dataset_type
        return self.sync_seqr_project_endpoint.call_with_http_info(**kwargs)




"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 7.4.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from metamist.api_client import ApiClient, Endpoint as _Endpoint
from metamist.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types,
    async_wrap
)
from metamist.model.http_validation_error import HTTPValidationError
from metamist.model.participant_upsert import ParticipantUpsert
from metamist.model.query_participant_criteria import QueryParticipantCriteria


class ParticipantApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        self.fill_in_missing_participants_async = async_wrap(self.fill_in_missing_participants)
        self.fill_in_missing_participants_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}/fill-in-missing-participants',
                'operation_id': 'fill_in_missing_participants',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_external_participant_id_to_sequencing_group_id_async = async_wrap(self.get_external_participant_id_to_sequencing_group_id)
        self.get_external_participant_id_to_sequencing_group_id_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}/external-pid-to-sg-id',
                'operation_id': 'get_external_participant_id_to_sequencing_group_id',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'sequencing_type',
                    'export_type',
                    'flip_columns',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'sequencing_type':
                        (str,),
                    'export_type':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                    'flip_columns':
                        (bool,),
                },
                'attribute_map': {
                    'project': 'project',
                    'sequencing_type': 'sequencing_type',
                    'export_type': 'export_type',
                    'flip_columns': 'flip_columns',
                },
                'location_map': {
                    'project': 'path',
                    'sequencing_type': 'query',
                    'export_type': 'query',
                    'flip_columns': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_individual_metadata_for_seqr_async = async_wrap(self.get_individual_metadata_for_seqr)
        self.get_individual_metadata_for_seqr_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}/individual-metadata-seqr',
                'operation_id': 'get_individual_metadata_for_seqr',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'export_type',
                    'external_participant_ids',
                    'replace_with_participant_external_ids',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                    'external_participant_ids',
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'export_type':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                    'external_participant_ids':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                    'replace_with_participant_external_ids':
                        (bool,),
                },
                'attribute_map': {
                    'project': 'project',
                    'export_type': 'export_type',
                    'external_participant_ids': 'external_participant_ids',
                    'replace_with_participant_external_ids': 'replace_with_participant_external_ids',
                },
                'location_map': {
                    'project': 'path',
                    'export_type': 'query',
                    'external_participant_ids': 'query',
                    'replace_with_participant_external_ids': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_participant_id_map_by_external_ids_async = async_wrap(self.get_participant_id_map_by_external_ids)
        self.get_participant_id_map_by_external_ids_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}/id-map/external',
                'operation_id': 'get_participant_id_map_by_external_ids',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'request_body',
                    'allow_missing',
                ],
                'required': [
                    'project',
                    'request_body',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'request_body':
                        ([str],),
                    'allow_missing':
                        (bool,),
                },
                'attribute_map': {
                    'project': 'project',
                    'allow_missing': 'allow_missing',
                },
                'location_map': {
                    'project': 'path',
                    'request_body': 'body',
                    'allow_missing': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.get_participants_async = async_wrap(self.get_participants)
        self.get_participants_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}',
                'operation_id': 'get_participants',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'query_participant_criteria',
                ],
                'required': [
                    'project',
                    'query_participant_criteria',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'query_participant_criteria':
                        (QueryParticipantCriteria,),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                    'query_participant_criteria': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.update_many_participants_async = async_wrap(self.update_many_participants)
        self.update_many_participants_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/update-many',
                'operation_id': 'update_many_participants',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'request_body',
                ],
                'required': [
                    'request_body',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'request_body':
                        ({str: (str,)},),
                },
                'attribute_map': {
                },
                'location_map': {
                    'request_body': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.update_participant_async = async_wrap(self.update_participant)
        self.update_participant_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{participant_id}/update-participant',
                'operation_id': 'update_participant',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'participant_id',
                    'participant_upsert',
                ],
                'required': [
                    'participant_id',
                    'participant_upsert',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'participant_id':
                        (int,),
                    'participant_upsert':
                        (ParticipantUpsert,),
                },
                'attribute_map': {
                    'participant_id': 'participant_id',
                },
                'location_map': {
                    'participant_id': 'path',
                    'participant_upsert': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.update_participant_family_async = async_wrap(self.update_participant_family)
        self.update_participant_family_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{participant_id}/update-participant-family',
                'operation_id': 'update_participant_family',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'participant_id',
                    'old_family_id',
                    'new_family_id',
                ],
                'required': [
                    'participant_id',
                    'old_family_id',
                    'new_family_id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'participant_id':
                        (int,),
                    'old_family_id':
                        (int,),
                    'new_family_id':
                        (int,),
                },
                'attribute_map': {
                    'participant_id': 'participant_id',
                    'old_family_id': 'old_family_id',
                    'new_family_id': 'new_family_id',
                },
                'location_map': {
                    'participant_id': 'path',
                    'old_family_id': 'query',
                    'new_family_id': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.upsert_participants_async = async_wrap(self.upsert_participants)
        self.upsert_participants_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/participant/{project}/upsert-many',
                'operation_id': 'upsert_participants',
                'http_method': 'PUT',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'participant_upsert',
                ],
                'required': [
                    'project',
                    'participant_upsert',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'participant_upsert':
                        ([ParticipantUpsert],),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                    'participant_upsert': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

    def fill_in_missing_participants(
        self,
        project,
        **kwargs
    ):
        """Fill In Missing Participants  # noqa: E501

        Create a corresponding participant (if required) for each sample within a project, useful for then importing a pedigree  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.fill_in_missing_participants(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.fill_in_missing_participants_endpoint.call_with_http_info(**kwargs)

    def get_external_participant_id_to_sequencing_group_id(
        self,
        project,
        **kwargs
    ):
        """Get External Participant Id To Sequencing Group Id  # noqa: E501

        Get csv / tsv export of external_participant_id to sequencing_group_id  Get a map of {external_participant_id} -> {sequencing_group_id} useful to matching joint-called sequencing groups in the matrix table to the participant  Return a list not dictionary, because dict could lose participants with multiple samples.  :param sequencing_type: Leave empty to get all sequencing types :param flip_columns: Set to True when exporting for seqr  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_external_participant_id_to_sequencing_group_id(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            sequencing_type (str): [optional]
            export_type (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            flip_columns (bool): [optional] if omitted the server will use the default value of False
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.get_external_participant_id_to_sequencing_group_id_endpoint.call_with_http_info(**kwargs)

    def get_individual_metadata_for_seqr(
        self,
        project,
        **kwargs
    ):
        """Get Individual Metadata Template For Seqr  # noqa: E501

        Get individual metadata template for SEQR as a CSV  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_individual_metadata_for_seqr(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            export_type (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            external_participant_ids (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            replace_with_participant_external_ids (bool): [optional] if omitted the server will use the default value of True
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.get_individual_metadata_for_seqr_endpoint.call_with_http_info(**kwargs)

    def get_participant_id_map_by_external_ids(
        self,
        project,
        request_body,
        **kwargs
    ):
        """Get Id Map By External Ids  # noqa: E501

        Get ID map of participants, by external_id  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_participant_id_map_by_external_ids(project, request_body, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            request_body ([str]):

        Keyword Args:
            allow_missing (bool): [optional] if omitted the server will use the default value of False
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['request_body'] = \
            request_body
        return self.get_participant_id_map_by_external_ids_endpoint.call_with_http_info(**kwargs)

    def get_participants(
        self,
        project,
        query_participant_criteria,
        **kwargs
    ):
        """Get Participants  # noqa: E501

        Get participants, default ALL participants in project  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_participants(project, query_participant_criteria, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            query_participant_criteria (QueryParticipantCriteria):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['query_participant_criteria'] = \
            query_participant_criteria
        return self.get_participants_endpoint.call_with_http_info(**kwargs)

    def update_many_participants(
        self,
        request_body,
        **kwargs
    ):
        """Update Many Participant External Ids  # noqa: E501

        Update external_ids of participants by providing an update map  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_many_participants(request_body, async_req=True)
        >>> result = thread.get()

        Args:
            request_body ({str: (str,)}):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['request_body'] = \
            request_body
        return self.update_many_participants_endpoint.call_with_http_info(**kwargs)

    def update_participant(
        self,
        participant_id,
        participant_upsert,
        **kwargs
    ):
        """Update Participant  # noqa: E501

        Update Participant Data  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_participant(participant_id, participant_upsert, async_req=True)
        >>> result = thread.get()

        Args:
            participant_id (int):
            participant_upsert (ParticipantUpsert):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['participant_id'] = \
            participant_id
        kwargs['participant_upsert'] = \
            participant_upsert
        return self.update_participant_endpoint.call_with_http_info(**kwargs)

    def update_participant_family(
        self,
        participant_id,
        old_family_id,
        new_family_id,
        **kwargs
    ):
        """Update Participant Family  # noqa: E501

        Change a participants family from old_family_id to new_family_id, maintaining all other fields. The new_family_id must already exist.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_participant_family(participant_id, old_family_id, new_family_id, async_req=True)
        >>> result = thread.get()

        Args:
            participant_id (int):
            old_family_id (int):
            new_family_id (int):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['participant_id'] = \
            participant_id
        kwargs['old_family_id'] = \
            old_family_id
        kwargs['new_family_id'] = \
            new_family_id
        return self.update_participant_family_endpoint.call_with_http_info(**kwargs)

    def upsert_participants(
        self,
        project,
        participant_upsert,
        **kwargs
    ):
        """Upsert Participants  # noqa: E501

        Upserts a list of participants with samples and sequences Returns the list of internal sample IDs  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.upsert_participants(project, participant_upsert, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            participant_upsert ([ParticipantUpsert]):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['participant_upsert'] = \
            participant_upsert
        return self.upsert_participants_endpoint.call_with_http_info(**kwargs)




from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

import api.routes.cohort
import metamist.models
from models.models.cohort import CohortBody, CohortCriteria, NewCohort
from models.utils.cohort_template_id_format import cohort_template_id_format
from scripts.create_custom_cohort import get_cohort_spec, main


class TestCohortBuilder(DbIsolatedTest):
    """Test custom cohort builder script"""

    @run_as_sync
    async def setUp(self):
        super().setUp()

    @run_as_sync
    async def mock_ccfc(self, project, body_create_cohort_from_criteria):
        """Mock by directly calling the API route"""
        self.assertEqual(project, self.project_name)
        return await api.routes.cohort.create_cohort_from_criteria(
            CohortBody(**body_create_cohort_from_criteria['cohort_spec'].to_dict()),
            CohortCriteria(
                **body_create_cohort_from_criteria['cohort_criteria'].to_dict()
            ),
            self.connection,
            body_create_cohort_from_criteria['dry_run'],
        )

    @run_as_sync
    async def test_get_cohort_spec(self):
        """Test get_cohort_spec(), invoked by the creator script"""
        ctemplate_id = cohort_template_id_format(28)
        result = get_cohort_spec('My cohort', 'Describing the cohort', ctemplate_id)
        self.assertIsInstance(result, metamist.models.CohortBody)
        self.assertEqual(result.name, 'My cohort')
        self.assertEqual(result.description, 'Describing the cohort')
        self.assertEqual(result.template_id, ctemplate_id)

    @run_as_sync
    @patch('metamist.apis.CohortApi.create_cohort_from_criteria')
    async def test_empty_main(self, mock):
        """Test main with no criteria"""
        mock.side_effect = self.mock_ccfc
        result = main(
            project=self.project_name,
            cohort_body_spec=metamist.models.CohortBody(
                name='Empty cohort', description='No criteria'
            ),
            projects=['test'],
            sg_ids_internal=[],
            excluded_sg_ids=[],
            sg_technologies=[],
            sg_platforms=[],
            sg_types=[],
            sample_types=[],
            dry_run=False,
        )
        mock.assert_called_once()
        self.assertIsInstance(result, NewCohort)
        self.assertIsInstance(result.cohort_id, str)
        self.assertListEqual(result.sequencing_group_ids, [])
        self.assertEqual(result.dry_run, False)

    @run_as_sync
    @patch('metamist.apis.CohortApi.create_cohort_from_criteria')
    async def test_epic_main(self, mock):
        """Test"""
        mock.side_effect = self.mock_ccfc
        result = main(
            project=self.project_name,
            cohort_body_spec=metamist.models.CohortBody(
                name='Epic cohort', description='Every criterion'
            ),
            projects=['test'],
            sg_ids_internal=['CPGLCL33'],
            excluded_sg_ids=['CPGLCL17', 'CPGLCL25'],
            sg_technologies=['short-read'],
            sg_platforms=['illumina'],
            sg_types=['genome'],
            sample_types=['blood'],
            dry_run=False,
        )
        mock.assert_called_once()
        self.assertEqual(mock.call_args.kwargs['project'], self.project_name)

        body = mock.call_args.kwargs['body_create_cohort_from_criteria']
        spec = body['cohort_spec']
        self.assertEqual(spec.name, 'Epic cohort')
        self.assertEqual(spec.description, 'Every criterion')

        criteria = body['cohort_criteria']
        self.assertIsInstance(criteria, metamist.models.CohortCriteria)
        self.assertListEqual(criteria.projects, ['test'])
        self.assertListEqual(criteria.sg_ids_internal, ['CPGLCL33'])
        self.assertListEqual(criteria.excluded_sgs_internal, ['CPGLCL17', 'CPGLCL25'])
        self.assertListEqual(criteria.sg_technology, ['short-read'])
        self.assertListEqual(criteria.sg_platform, ['illumina'])
        self.assertListEqual(criteria.sg_type, ['genome'])
        self.assertListEqual(criteria.sample_type, ['blood'])

        self.assertFalse(body['dry_run'])

        self.assertIsInstance(result, NewCohort)
        self.assertListEqual(result.sequencing_group_ids, [])
        self.assertEqual(result.dry_run, False)



#!/usr/bin/env python
"""
Report assay files in the main-upload bucket that can be deleted, ingested,
and sequencing groups that have no aligned CRAM.
"""

import asyncio
import logging
import os
import sys
from datetime import datetime
from functools import cache

import click

from cpg_utils.config import get_config

from metamist.audit.generic_auditor import GenericAuditor
from metamist.graphql import gql, query

FASTQ_EXTENSIONS = ('.fq.gz', '.fastq.gz', '.fq', '.fastq')
BAM_EXTENSIONS = ('.bam',)
CRAM_EXTENSIONS = ('.cram',)
READ_EXTENSIONS = FASTQ_EXTENSIONS + BAM_EXTENSIONS + CRAM_EXTENSIONS
GVCF_EXTENSIONS = ('.g.vcf.gz',)
VCF_EXTENSIONS = ('.vcf', '.vcf.gz')
ALL_EXTENSIONS = (
    FASTQ_EXTENSIONS
    + BAM_EXTENSIONS
    + CRAM_EXTENSIONS
    + GVCF_EXTENSIONS
    + VCF_EXTENSIONS
)

FILE_TYPES_MAP = {
    'fastq': FASTQ_EXTENSIONS,
    'bam': BAM_EXTENSIONS,
    'cram': CRAM_EXTENSIONS,
    'all_reads': READ_EXTENSIONS,
    'gvcf': GVCF_EXTENSIONS,
    'vcf': VCF_EXTENSIONS,
    'all': ALL_EXTENSIONS,
}

SEQUENCING_TYPES_QUERY = gql(
    """
    query seqTypes {
        enum {
            sequencingType
        }
    }
    """
)


@cache
def get_sequencing_types():
    """Return the list of sequencing types from the enum table."""
    logging.getLogger().setLevel(logging.WARN)
    sequencing_types = query(SEQUENCING_TYPES_QUERY)
    logging.getLogger().setLevel(logging.INFO)
    return sequencing_types['enum'][  # pylint: disable=unsubscriptable-object
        'sequencingType'
    ]


def audit_upload_bucket(
    dataset: str,
    sequencing_types: list[str],
    file_types: list[str],
    default_analysis_type: str,
    default_analysis_status: str,
):
    """Entrypoint for running upload bucket auditor asynchronously."""
    asyncio.get_event_loop().run_until_complete(
        asyncio.wait_for(
            audit_upload_bucket_async(
                dataset,
                sequencing_types,
                file_types,
                default_analysis_type,
                default_analysis_status,
            ),
            timeout=60,
        )
    )


class UploadBucketAuditor(GenericAuditor):
    """Auditor for upload cloud storage buckets"""

    def __init__(
        self,
        dataset: str,
        sequencing_types: list[str],
        file_types: tuple[str],
        default_analysis_type='cram',
        default_analysis_status='completed',
    ):
        super().__init__(
            dataset=dataset,
            sequencing_types=sequencing_types,
            file_types=file_types,
            default_analysis_type=default_analysis_type,
            default_analysis_status=default_analysis_status,
        )

    async def write_upload_bucket_audit_reports(
        self,
        bucket_name: str,
        sequencing_types: list[str],
        file_types: list[str],
        assay_files_to_delete: list[tuple[str, int, str, list[int]]],
        assay_files_to_ingest: list[tuple[str, str, str, int, str]],
        unaligned_sgs: list[tuple[str, str]],
    ):
        """
        Writes the 'assay files to delete/ingest' csv reports and upload them to the bucket.
        Also writes a report for any assay files found that match existing samples and may
        require ingestion.

        The report names include the file types, sequencing types, and date of the audit.
        """
        today = datetime.today().strftime('%Y-%m-%d')

        report_path = f'gs://{bucket_name}/audit_results/{today}/'

        if set(sequencing_types) == set(get_sequencing_types()):
            sequencing_types_str = 'all'
        else:
            sequencing_types_str = ('_').join(sequencing_types)

        if set(file_types) == set(ALL_EXTENSIONS):
            file_types_str = 'all'
        elif set(file_types) == set(READ_EXTENSIONS):
            file_types_str = 'all_reads'
        else:
            file_types_str = ('_').join(file_types)

        report_prefix = f'{self.dataset}_{file_types_str}_{sequencing_types_str}'

        if not assay_files_to_delete:
            logging.info('No assay read files to delete found. Skipping report...')
        else:
            assays_to_delete_file = f'{report_prefix}_assay_files_to_delete_{today}.csv'
            self.write_csv_report_to_cloud(
                data_to_write=assay_files_to_delete,
                report_path=os.path.join(report_path, assays_to_delete_file),
                header_row=[
                    'SG_ID',
                    'Assay_ID',
                    'Assay_Read_File_Path',
                    'CRAM_Analysis_ID',
                    'Filesize',
                ],
            )

        # 'Sequences to ingest' report contains paths to the (possibly) uningested files - and any samples/SGs that might be related
        if not assay_files_to_ingest:
            logging.info('No assay reads to ingest found. Skipping report...')
        else:
            assays_to_ingest_file = f'{report_prefix}_assay_files_to_ingest_{today}.csv'
            self.write_csv_report_to_cloud(
                data_to_write=assay_files_to_ingest,
                report_path=os.path.join(report_path, assays_to_ingest_file),
                header_row=[
                    'Assay_File_Path',
                    'SG_ID',
                    'Sample_ID',
                    'Sample_External_ID',
                    'CRAM_Analysis_ID',
                    'CRAM_Path',
                ],
            )

        # Write the sequencing groups without any completed cram to a csv
        if not unaligned_sgs:
            logging.info('No sequencing groups without crams found. Skipping report...')
        else:
            unaligned_sgs_file = f'{report_prefix}_unaligned_sgs_{today}.csv'
            self.write_csv_report_to_cloud(
                data_to_write=unaligned_sgs,
                report_path=os.path.join(report_path, unaligned_sgs_file),
                header_row=['SG_ID', 'Sample_ID', 'Sample_External_ID'],
            )


async def audit_upload_bucket_async(
    dataset: str,
    sequencing_types: list[str],
    file_types: list[str],
    default_analysis_type: str,
    default_analysis_status: str,
):
    """
    Finds sequence files for samples with completed CRAMs and adds these to a csv for deletion.
    Also finds any extra files in the upload bucket which may be uningested sequence data.
    Reports any files to delete, files to ingest, and samples without completed crams in output files

    Arguments:
        dataset: The name of the dataset to audit
        sequencing_types: The list of sequencing types to audit
        file_types: The list of file types to audit
        default_analysis_type: The default analysis type to audit
        default_analysis_status: The default analysis status to audit
    """

    # Validate user inputs
    allowed_sequencing_types = get_sequencing_types()
    if sequencing_types != ('all',) and any(
        st not in allowed_sequencing_types for st in sequencing_types
    ):
        raise ValueError(
            f'Input sequencing types "{sequencing_types}" must be in the allowed types: {allowed_sequencing_types}'
        )
    if sequencing_types == ('all',):
        sequencing_types = allowed_sequencing_types

    if file_types not in (('all',), ('all_reads',)):
        if any(ft not in FILE_TYPES_MAP for ft in file_types):
            raise ValueError(
                f'Input file types "{file_types}" must be in the allowed types {(", ").join(list(FILE_TYPES_MAP.keys()))}'
            )
    else:
        file_types = FILE_TYPES_MAP[file_types[0]]

    config = get_config()
    if not dataset:
        dataset = config['workflow']['dataset']
    bucket = config['storage'][dataset]['upload']

    # Initialise the auditor
    auditor = UploadBucketAuditor(
        dataset=dataset,
        sequencing_types=sequencing_types,
        file_types=file_types,
        default_analysis_type=default_analysis_type,
        default_analysis_status=default_analysis_status,
    )

    participant_data = await auditor.get_participant_data_for_dataset()
    sample_internal_external_id_map = auditor.map_internal_to_external_sample_ids(
        participant_data
    )
    (
        sg_sample_id_map,
        assay_sg_id_map,
        assay_filepaths_filesizes,
    ) = auditor.get_assay_map_from_participants(participant_data)

    # Get all completed cram output paths for the samples in the dataset and validate them
    sg_cram_paths = await auditor.get_analysis_cram_paths_for_dataset_sgs(
        assay_sg_id_map
    )

    # Identify sgs with and without completed crams
    sg_completion = await auditor.get_complete_and_incomplete_sgs(
        assay_sg_id_map, sg_cram_paths
    )

    unaligned_sgs = [
        (
            sg_id,
            sg_sample_id_map[sg_id],
            sample_internal_external_id_map.get(sg_sample_id_map[sg_id]),
        )
        for sg_id in sg_completion.get('incomplete')
    ]

    (
        reads_to_delete,
        reads_to_ingest,
    ) = await auditor.get_reads_to_delete_or_ingest(
        bucket,
        sg_completion.get('complete'),
        assay_filepaths_filesizes,
        sg_sample_id_map,
        assay_sg_id_map,
        sample_internal_external_id_map,
    )

    possible_assay_ingests = auditor.find_crams_for_reads_to_ingest(
        reads_to_ingest, sg_cram_paths
    )

    # Write the reads to delete, reads to ingest, and unaligned SGs reports
    await auditor.write_upload_bucket_audit_reports(
        bucket,
        sequencing_types=sequencing_types,
        file_types=file_types,
        assay_files_to_delete=reads_to_delete,
        assay_files_to_ingest=possible_assay_ingests,
        unaligned_sgs=unaligned_sgs,
    )


@click.command()
@click.option(
    '--dataset',
    '-d',
    help='Metamist dataset, used to filter samples',
)
@click.option(
    '--sequencing-types',
    '-s',
    multiple=True,
    required=True,
    help=f'"all", or any of {", ".join(get_sequencing_types())}',
)
@click.option(
    '--file-types',
    '-f',
    multiple=True,
    required='True',
    help='Find fastq, bam, cram, gvcf, vcf, all_reads (fastq + bam + cram), or all sequence file types',
)
def main(
    dataset: str,
    sequencing_types: tuple[str],
    file_types: tuple[str],
    default_analysis_type='cram',
    default_analysis_status='completed',
):
    """Runs the auditor on the dataset"""
    audit_upload_bucket(
        dataset,
        sequencing_types,
        file_types,
        default_analysis_type,
        default_analysis_status,
    )


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(levelname)s %(module)s:%(lineno)d - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        stream=sys.stderr,
    )
    main()  # pylint: disable=no-value-for-parameter



import unittest
from io import StringIO
from test.testbase import run_as_sync
from unittest.mock import patch

from scripts.process_ont_products import OntProductParser


class TestOntSampleSheetParser(unittest.TestCase):
    """Test the TestOntSampleSheetParser"""

    @run_as_sync
    @patch('metamist.apis.SampleApi.get_sample_id_map_by_external')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_single_row_all_files_exist(
        self, mock_filesize, mock_fileexists, mock_get_sample_id
    ):
        """
        Test processing one row with all files existing
        """
        mock_get_sample_id.return_value = {'Sample01': 'CPGaaa'}
        mock_filesize.return_value = 111
        mock_fileexists.return_value = True

        rows = [
            'Experiment name,Sample ID,Alignment_file,Alignment_software,SV_file,SV_software,SNV_file,SNV_software,Indel_file,Indel_software',
            'PBX10,Sample01,Sample01.bam,minimap2/2.22,Sample01.sv.vcf.gz,"Sniffles2, Version 2.0.2",Sample01.snvs.vcf.gz,Clair3 v0.1-r7,Sample01.indels.vcf.gz,Clair3 v0.1-r7',
        ]

        parser = OntProductParser(
            search_paths=[],
            # doesn't matter, we're going to mock the call anyway
            project='dev',
            dry_run=True,
        )

        # parser.skip_checking_gcs_objects = True
        fs = [
            'Sample01.bam',
            'Sample01.sv.vcf.gz',
            'Sample01.snvs.vcf.gz',
            'Sample01.indels.vcf.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        # parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        analyses = await parser.parse_manifest(
            StringIO(file_contents),
            delimiter=',',
        )

        self.assertEqual(4, len(analyses))



import unittest
from datetime import datetime
from io import StringIO
from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

import api.graphql.schema
from db.python.layers import ParticipantLayer
from metamist.graphql import configure_sync_client, validate
from metamist.parser.generic_metadata_parser import (
    DefaultSequencing,
    GenericMetadataParser,
)
from metamist.parser.generic_parser import (
    QUERY_MATCH_ASSAYS,
    QUERY_MATCH_PARTICIPANTS,
    QUERY_MATCH_SAMPLES,
    QUERY_MATCH_SEQUENCING_GROUPS,
    ParsedParticipant,
    ParsedSample,
    ParsedSequencingGroup,
)
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.utils.sample_id_format import sample_id_format
from models.utils.sequencing_group_id_format import sequencing_group_id_format


def _get_basic_participant_to_upsert():
    default_assay_meta = {
        'sequencing_type': 'genome',
        'sequencing_technology': 'short-read',
        'sequencing_platform': 'illumina',
    }

    return ParticipantUpsertInternal(
        external_ids={PRIMARY_EXTERNAL_ORG: 'Demeter'},
        meta={},
        samples=[
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'sample_id001'},
                meta={},
                type='blood',
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'reads': [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                    **default_assay_meta,
                                },
                            ),
                        ],
                    )
                ],
            )
        ],
    )


class TestValidateParserQueries(unittest.TestCase):
    """
    Validate queries used by the GenericParser
    """

    def test_queries(self):
        """
        We have the luxury of getting the schema directly, so we can validate
        the current development version! Oustide metamist, you'd just leave the
        schema option blank, and it would fetch the schema from the server.
        """

        # only need to apply schema to the first client to create, then it gets cached
        client = configure_sync_client(
            schema=api.graphql.schema.schema.as_str(),
            auth_token='FAKE',  # type: ignore
        )
        validate(QUERY_MATCH_PARTICIPANTS)
        validate(QUERY_MATCH_SAMPLES, client=client)
        validate(QUERY_MATCH_SEQUENCING_GROUPS, client=client)
        validate(QUERY_MATCH_ASSAYS, client=client)


class TestParseGenericMetadata(DbIsolatedTest):
    """Test the GenericMetadataParser"""

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('os.path.getsize')
    async def test_key_map(self, mock_stat_size, mock_graphql_query):
        """
        Test the flexible key map + other options
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async
        mock_stat_size.return_value = 111

        rows = [
            'sample id,filenames',
            '<sample-id>,<sample-id>-R1.fastq.gz',
            '<sample-id>,<sample-id>-R2.fastq.gz',
        ]

        parser = GenericMetadataParser(
            search_locations=['.'],
            key_map={
                'fn': ['filenames', 'filename'],
                'sample': ['sample id', 'sample_id'],
            },
            ignore_extra_keys=False,
            reads_column='fn',
            sample_primary_eid_column='sample',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            '<sample-id>-R1.fastq.gz': 'gs://<sample-id>-R1.fastq.gz',
            '<sample-id>-R2.fastq.gz': 'gs://<sample-id>-R2.fastq.gz',
        }
        # samples: list[ParsedSample]
        summary, _ = await parser.parse_manifest(
            StringIO('\n'.join(rows)), delimiter=',', dry_run=True
        )

        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(0, summary.assays.update)

        parser.ignore_extra_keys = False
        rows = [
            'sample id,filenames,extra',
            '<sample-id>,<sample-id>-R1.fastq.gz,extra',
            '<sample-id>,<sample-id>-R2.fastq.gz,read-all-about-it',
        ]

        try:
            _ = await parser.parse_manifest(
                StringIO('\n'.join(rows)), delimiter=',', dry_run=True
            )
        except ValueError as e:
            self.assertEqual(
                "Key 'extra' not found in provided key map: fn, sample", str(e)
            )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.datetime_added')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_single_row(
        self, mock_filesize, mock_fileexists, mock_datetime_added, mock_graphql_query
    ):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_assay_ids_for_sample_ids_by_type
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_filesize.return_value = 111
        mock_fileexists.return_value = False
        mock_datetime_added.return_value = datetime.fromisoformat('2022-02-02T22:22:22')

        rows = [
            'GVCF\tCRAM\tSampleId\tsample.flowcell_lane\tsample.platform\tsample.centre\tsample.reference_genome\traw_data.FREEMIX\traw_data.PCT_CHIMERAS\traw_data.MEDIAN_INSERT_SIZE\traw_data.MEDIAN_COVERAGE',
            '<sample-id>.g.vcf.gz\t<sample-id>.bam\t<sample-id>\tHK7NFCCXX.1\tILLUMINA\tKCCG\thg38\t0.01\t0.01\t400\t30',
        ]
        parser = GenericMetadataParser(
            search_locations=[],
            sample_primary_eid_column='SampleId',
            participant_meta_map={},
            sample_meta_map={'sample.centre': 'centre'},
            assay_meta_map={
                'raw_data.FREEMIX': 'qc.freemix',
                'raw_data.PCT_CHIMERAS': 'qc.pct_chimeras',
                'raw_data.MEDIAN_INSERT_SIZE': 'qc.median_insert_size',
                'raw_data.MEDIAN_COVERAGE': 'qc.median_coverage',
            },
            qc_meta_map={
                'raw_data.FREEMIX': 'freemix',
                'raw_data.PCT_CHIMERAS': 'pct_chimeras',
                'raw_data.MEDIAN_INSERT_SIZE': 'median_insert_size',
                'raw_data.MEDIAN_COVERAGE': 'median_coverage',
            },
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            reads_column='CRAM',
            gvcf_column='GVCF',
        )

        parser.filename_map = {
            '<sample-id>.g.vcf.gz': '/path/to/<sample-id>.g.vcf.gz',
            '<sample-id>.bam': '/path/to/<sample-id>.bam',
        }

        file_contents = '\n'.join(rows)
        summary, samples = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(0, summary.assays.update)
        self.assertEqual(1, summary.analyses.insert)

        self.assertDictEqual({'centre': 'KCCG'}, samples[0].meta)
        expected_assay_dict = {
            'qc': {
                'median_insert_size': 400,
                'median_coverage': 30,
                'freemix': 0.01,
                'pct_chimeras': 0.01,
            },
            'reads_type': 'bam',
            'reads': {
                'location': '/path/to/<sample-id>.bam',
                'basename': '<sample-id>.bam',
                'class': 'File',
                'checksum': None,
                'size': 111,
                'datetime_added': '2022-02-02T22:22:22',
            },
            'sequencing_platform': 'illumina',
            'sequencing_technology': 'short-read',
            'sequencing_type': 'genome',
        }

        assay_group_dict = {
            'gvcf_types': 'gvcf',
            'gvcfs': [
                {
                    'location': '/path/to/<sample-id>.g.vcf.gz',
                    'basename': '<sample-id>.g.vcf.gz',
                    'class': 'File',
                    'checksum': None,
                    'size': 111,
                    'datetime_added': '2022-02-02T22:22:22',
                }
            ],
        }
        self.assertDictEqual(assay_group_dict, samples[0].sequencing_groups[0].meta)
        self.assertDictEqual(
            expected_assay_dict, samples[0].sequencing_groups[0].assays[0].meta
        )
        analysis = samples[0].sequencing_groups[0].analyses[0]
        self.assertDictEqual(
            {
                'median_insert_size': 400,
                'median_coverage': 30,
                'freemix': 0.01,
                'pct_chimeras': 0.01,
            },
            analysis.meta,
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_rows_with_participants(self, mock_graphql_query):
        """
        Test importing a single row with a participant id, forms objects and checks response
        - MOCKS: query_async
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tSample ID\tFilenames\tType',
            'Demeter\tsample_id001\tsample_id001.filename-R1.fastq.gz,sample_id001.filename-R2.fastq.gz\tWGS',
            'Demeter\tsample_id001\tsample_id001.exome.filename-R1.fastq.gz,sample_id001.exome.filename-R2.fastq.gz\tWES',
            'Apollo\tsample_id002\tsample_id002.filename-R1.fastq.gz\tWGS',
            'Apollo\tsample_id002\tsample_id002.filename-R2.fastq.gz\tWGS',
            'Athena\tsample_id003\tsample_id003.filename-R1.fastq.gz',
            'Athena\tsample_id003\tsample_id003.filename-R2.fastq.gz',
            'Apollo\tsample_id004\tsample_id004.filename-R1.fastq.gz',
            'Apollo\tsample_id004\tsample_id004.filename-R2.fastq.gz',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_primary_eid_column='Individual ID',
            sample_primary_eid_column='Sample ID',
            reads_column='Filenames',
            seq_type_column='Type',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )

        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'sample_id001.filename-R1.fastq.gz': '/path/to/sample_id001.filename-R1.fastq.gz',
            'sample_id001.filename-R2.fastq.gz': '/path/to/sample_id001.filename-R2.fastq.gz',
            'sample_id001.exome.filename-R1.fastq.gz': '/path/to/sample_id001.exome.filename-R1.fastq.gz',
            'sample_id001.exome.filename-R2.fastq.gz': '/path/to/sample_id001.exome.filename-R2.fastq.gz',
            'sample_id002.filename-R1.fastq.gz': '/path/to/sample_id002.filename-R1.fastq.gz',
            'sample_id002.filename-R2.fastq.gz': '/path/to/sample_id002.filename-R2.fastq.gz',
            'sample_id003.filename-R1.fastq.gz': '/path/to/sample_id003.filename-R1.fastq.gz',
            'sample_id003.filename-R2.fastq.gz': '/path/to/sample_id003.filename-R2.fastq.gz',
            'sample_id004.filename-R1.fastq.gz': '/path/to/sample_id004.filename-R1.fastq.gz',
            'sample_id004.filename-R2.fastq.gz': '/path/to/sample_id004.filename-R2.fastq.gz',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        summary, prows = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        participants: list[ParsedParticipant] = prows

        self.assertEqual(3, summary.participants.insert)
        self.assertEqual(0, summary.participants.update)
        self.assertEqual(4, summary.samples.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(5, summary.assays.insert)
        self.assertEqual(0, summary.assays.update)
        self.assertEqual(0, summary.analyses.insert)

        expected_assay_dict = {
            'reads': [
                {
                    'basename': 'sample_id001.filename-R1.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
                {
                    'basename': 'sample_id001.filename-R2.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
            ],
            'reads_type': 'fastq',
            'sequencing_platform': 'illumina',
            'sequencing_technology': 'short-read',
            'sequencing_type': 'genome',
        }
        assay = participants[0].samples[0].sequencing_groups[0].assays[0]
        self.maxDiff = None
        self.assertDictEqual(expected_assay_dict, assay.meta)

        # Check that both of Demeter's assays are there
        self.assertEqual(participants[0].primary_external_id, 'Demeter')
        self.assertEqual(len(participants[0].samples), 1)
        self.assertEqual(len(participants[0].samples[0].sequencing_groups), 2)

        return

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_rows_with_valid_participant_meta(self, mock_graphql_query):
        """
        Test importing a several rows with a participant metadata (reported gender, sex and karyotype),
        forms objects and checks response
        - MOCKS: get_sample_id_map_by_external,  get_participant_id_map_by_external_ids,
        get_assay_ids_for_sample_ids_by_type
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tSample ID\tSex\tGender\tKaryotype',
            'Demeter\tsample_id001\tMale\tNon-binary\tXY',
            'Apollo\tsample_id002\tFemale\tFemale\tXX',
            'Athena\tsample_id003\tFEMalE',
            'Dionysus\tsample_id00x\t\tMale\tXX',
            'Pluto\tsample_id00y',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_primary_eid_column='Individual ID',
            sample_primary_eid_column='Sample ID',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            reported_sex_column='Sex',
            reported_gender_column='Gender',
            karyotype_column='Karyotype',
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )

        # Call generic parser
        file_contents = '\n'.join(rows)
        participants: list[ParsedParticipant]
        _, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )
        p_by_name = {p.primary_external_id: p for p in participants}
        demeter = p_by_name['Demeter']
        apollo = p_by_name['Apollo']
        athena = p_by_name['Athena']
        dionysus = p_by_name['Dionysus']
        # pluto = p_by_name['Pluto']

        # Assert that the participant meta is there.
        self.assertEqual(demeter.reported_gender, 'Non-binary')
        self.assertEqual(demeter.reported_sex, 1)
        self.assertEqual(demeter.karyotype, 'XY')
        self.assertEqual(apollo.reported_gender, 'Female')
        self.assertEqual(apollo.reported_sex, 2)
        self.assertEqual(apollo.karyotype, 'XX')
        self.assertEqual(athena.reported_sex, 2)
        self.assertIsNone(athena.reported_gender)
        self.assertIsNone(athena.karyotype)
        self.assertEqual(dionysus.reported_gender, 'Male')
        self.assertEqual(dionysus.karyotype, 'XX')
        return

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_rows_with_invalid_participant_meta(self, mock_graphql_query):
        """
        Test importing a single rows with invalid participant metadata,
        forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_participant_id_map_by_external_ids
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tSample ID\tSex\tKaryotype',
            'Athena\tsample_id003\tFemalee\tXX',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_primary_eid_column='Individual ID',
            sample_primary_eid_column='Sample ID',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            reported_sex_column='Sex',
            karyotype_column='Karyotype',
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )

        # Call generic parser
        file_contents = '\n'.join(rows)
        with self.assertRaises(ValueError):
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        return

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_lrs_rows_with_arbitrary_assay_meta_columns(self, mock_graphql_query):
        """
        Test importing rows of long read sequencing data with arbitrary assay metadata columns
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async
        self.maxDiff = None
        rows = [
            'Individual ID\tSample ID\tFilename\tAssay Meta 1\tAssay Meta 2',
            'Athena\tsample_id003\t"sample_id003.filename-01-R1.fastq.gz,sample_id003.filename-01-R2.fastq.gz"\tTrue\tsome_value',
            'Athena\tsample_id003\t"sample_id003.filename-02-R1.fastq.gz,sample_id003.filename-02-R2.fastq.gz"\t5\tFalse',
        ]
        parser = GenericMetadataParser(
            search_locations=[],
            participant_primary_eid_column='Individual ID',
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            assay_meta_map={
                'Assay Meta 1': 'assay_meta1',
                'Assay Meta 2': 'assay_meta2',
            },
            default_sequencing=DefaultSequencing(
                seq_type='genome',
                technology='long-read',
                platform='pacbio',
            ),
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            skip_checking_gcs_objects=True,
        )

        parser.filename_map = {
            'sample_id003.filename-01-R1.fastq.gz': '/path/to/sample_id003.filename-01-R1.fastq.gz',
            'sample_id003.filename-01-R2.fastq.gz': '/path/to/sample_id003.filename-01-R2.fastq.gz',
            'sample_id003.filename-02-R1.fastq.gz': '/path/to/sample_id003.filename-02-R1.fastq.gz',
            'sample_id003.filename-02-R2.fastq.gz': '/path/to/sample_id003.filename-02-R2.fastq.gz',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)

        _, prows = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        participants: list[ParsedParticipant] = prows

        expected_assay_1_dict = {
            'reads': [
                {
                    'basename': 'sample_id003.filename-01-R1.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id003.filename-01-R1.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
                {
                    'basename': 'sample_id003.filename-01-R2.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id003.filename-01-R2.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
            ],
            'reads_type': 'fastq',
            'sequencing_platform': 'pacbio',
            'sequencing_technology': 'long-read',
            'sequencing_type': 'genome',
            'assay_meta1': [5, True],
            'assay_meta2': [False, 'some_value'],
        }

        expected_assay_2_dict = {
            'reads': [
                {
                    'basename': 'sample_id003.filename-02-R1.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id003.filename-02-R1.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
                {
                    'basename': 'sample_id003.filename-02-R2.fastq.gz',
                    'checksum': None,
                    'class': 'File',
                    'location': '/path/to/sample_id003.filename-02-R2.fastq.gz',
                    'size': None,
                    'datetime_added': None,
                },
            ],
            'reads_type': 'fastq',
            'sequencing_platform': 'pacbio',
            'sequencing_technology': 'long-read',
            'sequencing_type': 'genome',
            'assay_meta1': [5, True],
            'assay_meta2': [False, 'some_value'],
        }

        expected_sg_dict = {
            'assay_meta1': [5, True],
            'assay_meta2': [False, 'some_value'],
        }

        assay_1 = participants[0].samples[0].sequencing_groups[0].assays[0]
        assay_2 = participants[0].samples[0].sequencing_groups[0].assays[1]
        sg = participants[0].samples[0].sequencing_groups[0]

        self.assertDictEqual(expected_assay_1_dict, assay_1.meta)
        self.assertDictEqual(expected_assay_2_dict, assay_2.meta)
        self.assertDictEqual(expected_sg_dict, sg.meta)
        return

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_contents')
    async def test_cram_with_no_reference(
        self,
        mock_filecontents,
        mock_filesize,
        mock_fileexists,
        mock_graphql_query,
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_filecontents.return_value = 'testmd5'
        mock_filesize.return_value = 111
        mock_fileexists.return_value = True

        rows = [
            'Sample ID\tFilename',
            'sample_id003\tfile.cram',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            skip_checking_gcs_objects=True,
        )

        parser.filename_map = {'file.cram': 'gs://path/file.cram'}

        # Call generic parser
        file_contents = '\n'.join(rows)
        with self.assertRaises(ValueError) as ctx:
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        self.assertEqual(
            "Reads type for 'sample_id003' is CRAM, but a reference is not defined, please set the default reference assembly path",
            str(ctx.exception),
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_cram_with_default_reference(
        self, mock_filesize, mock_fileexists, mock_graphql_query
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_filesize.return_value = 111
        mock_fileexists.return_value = True

        rows = [
            'Sample ID\tFilename',
            'sample_id003\tfile.cram',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file.fasta': 'gs://path/file.fasta',
            'file.fasta.fai': 'gs://path/file.fasta.fai',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        samples: list[ParsedSample]
        _, samples = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        expected = {
            'location': 'gs://path/file.fasta',
            'basename': 'file.fasta',
            'class': 'File',
            'checksum': None,
            'size': None,
            'datetime_added': None,
            'secondaryFiles': [
                {
                    'location': 'gs://path/file.fasta.fai',
                    'basename': 'file.fasta.fai',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                    'datetime_added': None,
                }
            ],
        }

        self.assertDictEqual(
            expected,
            samples[0].sequencing_groups[0].assays[0].meta['reference_assembly'],
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    async def test_cram_with_row_level_reference(
        self, mock_fileexists, mock_graphql_query
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_fileexists.return_value = True

        rows = [
            'Sample ID\tFilename\tRef',
            'sample_id003\tfile.cram\tref.fa',
            'sample_id003\tfile2.cram\tref.fa',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            reference_assembly_location_column='Ref',
            # default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file2.cram': 'gs://path/file2.cram',
            'ref.fa': 'gs://path/ref.fa',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        samples: list[ParsedSample]
        _, samples = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        expected = {
            'location': 'gs://path/ref.fa',
            'basename': 'ref.fa',
            'class': 'File',
            'checksum': None,
            'size': None,
            'datetime_added': None,
            'secondaryFiles': [
                {
                    'location': 'gs://path/ref.fa.fai',
                    'basename': 'ref.fa.fai',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                    'datetime_added': None,
                }
            ],
        }

        self.assertDictEqual(
            expected,
            samples[0].sequencing_groups[0].assays[0].meta['reference_assembly'],
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    async def test_cram_with_multiple_row_level_references(
        self, mock_fileexists, mock_graphql_query
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_fileexists.return_value = True

        rows = [
            'Sample ID\tFilename\tRef',
            'sample_id003\tfile.cram\tref.fa',
            'sample_id003\tfile2.cram\tref2.fa',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            reference_assembly_location_column='Ref',
            # default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file2.cram': 'gs://path/file.cram',
            'ref.fa': 'gs://path/ref.fa',
            'ref2.fa': 'gs://path/ref2.fa',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)

        with self.assertRaises(ValueError) as ctx:
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        self.assertEqual(
            'Multiple reference assemblies were defined for sample_id003: ref.fa, ref2.fa',
            str(ctx.exception),
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.datetime_added')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_matching_sequencing_groups_and_assays(
        self, mock_filesize, mock_fileexists, mock_datetime_added, mock_graphql_query
    ):
        """Test basic import with data that exists in the database"""
        mock_graphql_query.side_effect = self.run_graphql_query_async
        mock_filesize.return_value = 111
        mock_fileexists.return_value = False
        mock_datetime_added.return_value = datetime.fromisoformat('2022-02-02T22:22:22')

        player = ParticipantLayer(self.connection)
        participant = await player.upsert_participant(
            _get_basic_participant_to_upsert()
        )

        filenames = [
            'sample_id001.filename-R1.fastq.gz',
            'sample_id001.filename-R2.fastq.gz',
        ]

        rows = [
            'Participant ID\tSample ID\tFilename',
            f'Demeter\tsample_id001\t{filenames[0]}',
            f'Demeter\tsample_id001\t{filenames[1]}',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_primary_eid_column='Participant ID',
            sample_primary_eid_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            assay_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )

        parser.filename_map = {f: '/path/to/' + f for f in filenames}

        summary, parsed_files = await parser.parse_manifest(
            StringIO('\n'.join(rows)), delimiter='\t', dry_run=True
        )

        self.assertEqual(1, summary.participants.update)
        self.assertEqual(1, summary.samples.update)
        self.assertEqual(1, summary.sequencing_groups.update)
        self.assertEqual(1, summary.assays.update)
        self.assertEqual(0, summary.participants.insert)
        self.assertEqual(0, summary.samples.insert)
        self.assertEqual(0, summary.sequencing_groups.insert)
        self.assertEqual(0, summary.assays.insert)

        parsed_p: ParsedParticipant = parsed_files[0]
        self.assertEqual(participant.id, parsed_p.internal_pid)
        self.assertEqual(
            sample_id_format(participant.samples[0].id),
            parsed_p.samples[0].internal_sid,
        )

        sg: SequencingGroupUpsertInternal = participant.samples[0].sequencing_groups[0]
        sg_parsed: ParsedSequencingGroup = (
            parsed_files[0].samples[0].sequencing_groups[0]
        )

        self.assertEqual(
            sequencing_group_id_format(sg.id), sg_parsed.internal_seqgroup_id
        )
        self.assertEqual(len(sg.assays), len(sg_parsed.assays))
        self.assertEqual(sg.assays[0].id, sg_parsed.assays[0].internal_id)


class FastqPairMatcher(unittest.TestCase):
    """Test Fastq pair matching logic explictly"""

    def test_simple(self):
        """Simple fastq pair matching case"""
        entries = [
            'gs://BUCKET/FAKE/<sample-id>.filename-R2.fastq.gz',
            'gs://BUCKET/FAKE/<sample-id>.filename-R1.fastq.gz',
        ]

        grouped = GenericMetadataParser.parse_fastqs_structure(entries)

        self.assertEqual(1, len(grouped))
        self.assertEqual(2, len(grouped[0]))
        self.assertListEqual(sorted(grouped[0]), grouped[0])

    def test_post_r_value_matcher(self):
        """Test entries with post R values, eg: R1_001.fastq"""
        entries = [
            'gs://BUCKET/FAKE/<sample-id>.filename-R2_001.fastq.gz',
            'gs://BUCKET/FAKE/<sample-id>.filename-R1_001.fastq.gz',
            'gs://BUCKET/FAKE/<sample-id>.filename-R1_002.fastq.gz',
            'gs://BUCKET/FAKE/<sample-id>.filename-R2_002.fastq.gz',
        ]

        grouped = GenericMetadataParser.parse_fastqs_structure(entries)

        self.assertEqual(2, len(grouped))
        self.assertEqual(2, len(grouped[0]))

        # check that the 002 got grouped together
        self.assertIn('002', grouped[1][0])
        self.assertIn('002', grouped[1][1])



# pylint: disable=protected-access
import datetime
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from typing import Any

from db.python.tables.bq.billing_daily_extended import BillingDailyExtendedTable
from db.python.tables.bq.billing_filter import BillingFilter
from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from db.python.utils import InternalError
from models.models import BillingColumn, BillingTotalCostQueryModel


class TestBillingGcpDailyTable(BqTest):
    """Test BillingRawTable and its methods"""

    def setUp(self):
        super().setUp()

        # setup table object
        self.table_obj = BillingDailyExtendedTable(self.connection)

    def test_query_to_partitioned_filter(self):
        """Test query to partitioned filter conversion"""

        # given
        start_date = '2023-01-01'
        end_date = '2024-01-01'
        filters: dict[BillingColumn, str | list[Any] | dict[Any, Any]] = {
            BillingColumn.TOPIC: 'TEST_TOPIC'
        }

        # expected
        expected_filter = BillingFilter(
            day=GenericBQFilter(
                gte=datetime.datetime(2023, 1, 1, 0, 0),
                lte=datetime.datetime(2024, 1, 1, 0, 0),
            ),
            topic=GenericBQFilter(eq='TEST_TOPIC'),
        )

        query = BillingTotalCostQueryModel(
            fields=[],  # not relevant for this test, but can't be null generally
            start_date=start_date,
            end_date=end_date,
            filters=filters,
        )
        filter_ = BillingDailyExtendedTable._query_to_partitioned_filter(query)

        # BillingFilter has __eq__ method, so we can compare them directly
        self.assertEqual(expected_filter, filter_)

    def test_error_no_connection(self):
        """Test No connection exception"""

        with self.assertRaises(InternalError) as context:
            BillingDailyExtendedTable(None)

        self.assertTrue(
            "No connection was provided to the table 'BillingDailyExtendedTable'"
            in str(context.exception)
        )

    def test_get_table_name(self):
        """Test get_table_name"""

        # table name is set in the class
        given_table_name = 'TEST_TABLE_NAME'

        # set table name
        self.table_obj.table_name = given_table_name

        # test get table name function
        table_name = self.table_obj.get_table_name()

        self.assertEqual(given_table_name, table_name)

    @run_as_sync
    async def test_get_extended_values_return_empty_list(self):
        """Test get_extended_values as empty list"""

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get table name function
        records = await self.table_obj.get_extended_values('dataset')

        self.assertEqual([], records)

    @run_as_sync
    async def test_get_extended_values_return_valid_list(self):
        """Test get_extended_values as list of 2 records"""

        # mock BigQuery result as list of 2 records
        self.bq_result.result.return_value = [
            {'dataset': 'DATA1'},
            {'dataset': 'DATA2'},
        ]

        # test get table name function
        records = await self.table_obj.get_extended_values('dataset')

        self.assertEqual(['DATA1', 'DATA2'], records)

    @run_as_sync
    async def test_get_extended_values_error(self):
        """Test get_extended_values return exception as invalid ext column"""

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get table name function
        with self.assertRaises(ValueError) as context:
            await self.table_obj.get_extended_values('rubish')

        self.assertTrue('Invalid field value' in str(context.exception))
